CLASS_NAMES = {
    "ftagTruthOriginLabel": [
        "Pileup",
        "Fake",
        "Primary",
        "FromB",
        "FromBC",
        "FromC",
        "FromTau",
        "OtherSecondary",
    ],
    "ftagTruthTypeLabel": [
        "NoTruth",
        "Other",
        "Pion",
        "Kaon",
        "Electron",
        "Muon",
    ],
    "ftagTruthSourceLabel": [
        "NoTruth",
        "NotSecondary",
        "HadronicInteraction",
        "StrangeMesonDecay",
        "StrangeBaryonDecay",
        "GammaConversion",
        "Other",
    ],
}
